;==============================================================================
function give_KADEX_header_type, n_geo
   
   h = {file_name: 'xxx', $
        file_version: 'xxx', $
        citation_version: 'xxx', $
        documentation_file_name: 'xxx', $
        n_geo: n_geo, $
        date: strarr(n_geo), $
        latitude: fltarr(n_geo), $
        longitude: fltarr(n_geo), $
        sza: fltarr(n_geo), $
        altitude_range: [-999.99, -999.99], $
        target_name: 'xxx' $
       }

   return, h
end

;==============================================================================
function give_KADEX_result_type
   
   r = {file_name: 'xxx', $
        geo_id: 'xxx', $
        date: 'xxx', $
        latitude: 0.0, $
        longitude: 0.0, $
        sza: 0.0, $
        dof: 0.0, $
        target_name: 'xxx', $
        los: ptr_new(), $
        visibility: ptr_new(), $
        altitude: ptr_new(), $
        pressure: ptr_new(), $
        temperature: ptr_new(), $
        target: ptr_new(), $
        target_noise_error: ptr_new(), $
        akm_diagonal: ptr_new(), $
        vert_resolution: ptr_new(), $
        diagnostics_id: 0, $
        diagnostics_index: 0L, $
        total_error: ptr_new() $
       }

   return, r
end

;==============================================================================
function give_KADEX_diagnostics_type, n_diag
   
   d = {file_name: 'xxx', $
        diagnostics_id: 0, $
        altitude: ptr_new(), $
        total_error: ptr_new(), $
        random_error_covariance: ptr_new(), $
        ak_matrix: ptr_new() $
       }

   return, d
end

;==============================================================================
function map_kadex_diagnostics, result, diagnostics

   for i_diag=0, n_elements(diagnostics)-1 do begin
      index = where(result.diagnostics_id eq diagnostics[i_diag].diagnostics_id, count)
      if(count ne 0) then result[index].diagnostics_index = i_diag else return, 1
   endfor

   return, 0
end

;==============================================================================
function version_to_number, version_string

   v = strsplit(version_string, '.', /extract)
   if(n_elements(v) ne 3) then begin
      print, 'file version is not correct'
      stop
   endif else begin
      for i=0, 2 do begin
         if(where(v[i] eq strcompress(indgen(10),/remo)) eq -1) then begin
            print, 'file version is not correct'
            stop
         endif
      endfor
      version_number = round(total(v*[100,10,1]))
   endelse

   return, version_number
end

;==============================================================================
function read_kadex_header, d, file
   
   on_ioerror, IO_ERROR

   routine = 'read_kadex_header()'

;--- no parameter file available or file is no string: pick file from browser
   if(n_elements(file) ne 1) then begin
      file = dialog_pickfile(/read)
   endif else if((size(file, /str)).type_name ne 'STRING') then begin
      file = dialog_pickfile(/read)
   endif

;--- usually files are gzipped, so the /compress keyword is a good idea...
   openr, unit, file, /get_lun, /compress
   
;--- $0 read file version
   iwarn = go_section(0, 0, 0, 0, unit=unit)
   version = 'xxx'
   readf, unit, version
   version_number = version_to_number(version)

;--- $0.1 file name
   iwarn = go_section(0, 1, 0, 0, unit=unit)
   file_name = 'xxx'
   readf, unit, file_name
   
;--- $0.1.1 data version for citation in publications
   if(version_number ge 4) then begin
      iwarn = go_section(0, 1, 1, 0, unit=unit)
      citation_version = 'xxx'
      readf, unit, citation_version
   endif else begin
      citation_version = ''
   endelse

;--- $0.2 documentation file name
   iwarn = go_section(0, 2, 0, 0, unit=unit)
   doc_name = 'xxx'
   readf, unit, doc_name

;--- $1.2 number of result profiles (= geolocations = times)
   iwarn = go_section(1, 2, 0, 0, unit=unit)
   n_geo = 0
   readf, unit, n_geo

;--- prepare and partly fill header structure
   d = give_KADEX_header_type(n_geo)
   d.file_version = version
   d.file_name = file_name
   d.citation_version = citation_version
   d.documentation_file_name = doc_name

;--- $1.3 date/time, lat, lon, SZA for geolocs
   iwarn = go_section(1, 3, 0, 0, unit=unit)
   tmp1 = 'xxx' & tmp2 = 0.0 & tmp3 = 0.0 & tmp4 = 0.0
   for i_geo=0, n_geo-1 do begin
      readf, unit, tmp1, tmp2, tmp3, tmp4, format='(A17,3(f9.3))'
      d.date[i_geo]      = strcompress(tmp1, /remo)
      d.latitude[i_geo]  = tmp2
      d.longitude[i_geo] = tmp3
      d.sza[i_geo]       = tmp4
   endfor

;--- $1.4 useful altitude_range
   iwarn = go_section(1, 4, 0, 0, unit=unit)
   tmp = [0.0, 0.0]
   readf, unit, tmp
   d.altitude_range = tmp

;--- $1.5 target name
   iwarn = go_section(1, 5, 0, 0, unit=unit)
   tmp = 'xxx'
   readf, unit, tmp
   d.target_name = tmp

   free_lun, unit
   return, 0

IO_ERROR:
   if(n_elements(unit) ne 0) then free_lun, unit
   print, routine+': unspecified IO-error occured while reading file '+file
   return, 1

end

;==============================================================================
function read_kadex_result, d, file, index=index

   on_ioerror, IO_ERROR

   routine = 'read_kadex_result()'

;--- no paramter file available or file is no string: pick file from browser
   if(n_elements(file) ne 1) then begin
      file = dialog_pickfile(/read)
   endif else if((size(file, /str)).type_name ne 'STRING') then begin
      file = dialog_pickfile(/read)
   endif

;--- usually files are gzipped, so the /compress keyword is a good idea...
   openr, unit, file, /get_lun, /compress
   
;--- $0 this routine is only valid for files of version 0.0.3
   iwarn = go_section(0, 0, 0, 0, unit=unit)
   version = 'xxx'
   readf, unit, version
   version_number = version_to_number(version)
   
;--- $0.1 file name
   iwarn = go_section(0, 1, 0, 0, unit=unit)
   file_name = 'xxx'
   readf, unit, file_name
   
;--- $1.2 number of available result profiles (= geolocations = times)
   iwarn = go_section(1, 2, 0, 0, unit=unit)
   n_geo = 0
   readf, unit, n_geo

;--- $1.5 target name
   iwarn = go_section(1, 5, 0, 0, unit=unit)
   target_name = 'xxx'
   readf, unit, target_name

;--- check whether all data or a subset of the data should be read
   if(n_elements(index) eq 0) then index = lindgen(n_geo)

;--- redefinition of n_geo: number of required geolocations
   n_geo = n_elements(index)

;--- get an array of result structure
   d = replicate(give_KADEX_result_type(), n_geo)
   
;--- fill array (beware: IDL indices are $-item indices minus one)
   for i_geo=0, n_geo-1 do begin

;--- spread entries
      d[i_geo].file_name   = file_name
      d[i_geo].target_name = target_name

;--- $2.x.1 geo_id (composed of orbit, date, time, format format ooooo_YYYYMMDDThhmmssZ)
      iwarn = go_section(2, index[i_geo]+1, 1, 0, unit=unit)
      tmp = 'xxx'
      readf, unit, tmp
      d[i_geo].geo_id = strcompress(tmp, /remo)
      d[i_geo].date   = strmid(d[i_geo].geo_id, 6)

;--- $2.x.2 latitude (-90...90), longitude (-180...180) [deg]
      iwarn = go_section(2, index[i_geo]+1, 2, 0, unit=unit)
      tmp = [0.0, 0.0]
      readf, unit, tmp
      d[i_geo].latitude  = tmp[0]
      d[i_geo].longitude = tmp[1]

;--- $2.x.3 solar zenith angle [deg]
      iwarn = go_section(2, index[i_geo]+1, 3, 0, unit=unit)
      tmp = 0.0
      readf, unit, tmp
      d[i_geo].sza  = tmp[0]

;--- $2.x.6 degrees of freedom
      iwarn = go_section(2, index[i_geo]+1, 6, 0, unit=unit)
      tmp = 0.0
      readf, unit, tmp
      d[i_geo].dof = tmp

;--- $2.x.7 corresponding diagnostics data id
      iwarn = go_section(2, index[i_geo]+1, 7, 0, unit=unit)
      tmp = 0
      readf, unit, tmp
      d[i_geo].diagnostics_id = tmp

;--- $2.x.8 # of measurement tangent altitudes actually used
      iwarn = go_section(2, index[i_geo]+1, 8, 0, unit=unit)
      n_tang = 0.0
      readf, unit, n_tang

;--- $2.x.9 measurement tangent altitudes actually used [km]
      iwarn = go_section(2, index[i_geo]+1, 9, 0, unit=unit)
      tmp = fltarr(n_tang)
      readf, unit, tmp
      d[i_geo].los = ptr_new(tmp)

;--- $2.x.10 number of gridpoints
      iwarn = go_section(2, index[i_geo]+1, 10, 0, unit=unit)
      n_grid = 0.0
      readf, unit, n_grid

;--- $2.x.11 result entries
      iwarn = go_section(2, index[i_geo]+1, 11, 0, unit=unit)
      tmp = fltarr(8, n_grid)
      readf, unit, tmp
      d[i_geo].visibility         = ptr_new(round(tmp[0,*]))
      d[i_geo].altitude           = ptr_new(tmp[1,*])
      d[i_geo].pressure           = ptr_new(tmp[2,*])
      d[i_geo].temperature        = ptr_new(tmp[3,*])
      d[i_geo].target             = ptr_new(tmp[4,*])
      d[i_geo].target_noise_error = ptr_new(tmp[5,*])
      d[i_geo].akm_diagonal       = ptr_new(tmp[6,*])
      d[i_geo].vert_resolution    = ptr_new(tmp[7,*])

   endfor

   free_lun, unit
   return, 0

IO_ERROR:
   if(n_elements(unit) ne 0) then free_lun, unit
   print, routine+': unspecified IO-error occured while reading file '+file
   return, 1

end

;==============================================================================
function read_kadex_diagnostics, diagnostics, result, file, diag_id=diag_id

   on_ioerror, IO_ERROR

   routine = 'read_kadex_diagnostics()'
   TRUE  = 1
   FALSE = 0

;--- no paramter file available or file is no string: pick file from browser
   if(n_elements(file) ne 1) then begin
      file = dialog_pickfile(/read)
   endif else if((size(file, /str)).type_name ne 'STRING') then begin
      file = dialog_pickfile(/read)
   endif

;--- usually files are gzipped, so the /compress keyword is a good idea...
   openr, unit, file, /get_lun, /compress

;--- this routine is only valid for files of version 0.0.3
   iwarn = go_section(0, 0, 0, 0, unit=unit)
   version = 'xxx'
   readf, unit, version
   if(strcompress(version,/remo) ne '0.0.3') then begin
      print, routine+': this routine is only valid for files of version 0.0.3'
   endif

;--- $0.1 file name
   iwarn = go_section(0, 1, 0, 0, unit=unit)
   file_name = 'xxx'
   readf, unit, file_name

;--- jump to the diagnostics section of the file
   iwarn = go_section(3, 0, 1, 0, unit=unit)
   if(iwarn ne 0) then begin
      n_diag = 0
   endif else begin
      n_diag = 0
      readf, unit, n_diag
   endelse

   if(n_diag gt 0) then begin

;--- check whether all data or a subset of the data should be read
      diag_id = result[uniq(result.diagnostics_id, sort(result.diagnostics_id))].diagnostics_id
      n_to_read = n_elements(diag_id)

;--- get an array of diagnostics structure
      diagnostics = replicate(give_KADEX_diagnostics_type(), n_to_read)

;--- loop: jump to all diagnostics data blocks
      i_to_read = 0
      for i_diag=0, n_diag-1 do begin

;--- $3.x.1 diagnostics data id
         iwarn = go_section(3, i_diag+1, 1, 0, unit=unit)
         tmp = 0
         readf, unit, tmp

;--- if diagnostics id is not required: skip data block
         dummy = where(tmp eq diag_id, count)
         if(count eq 0) then continue

         diagnostics[i_to_read].file_name = file_name

;--- $3.x.1 diagnostics data id is copied into data structure
         diagnostics[i_to_read].diagnostics_id = tmp

;--- $3.x.2 number of altitude grid points
         iwarn = go_section(3, i_diag+1, 2, 0, unit=unit)
         n_grid = 0
         readf, unit, n_grid

;--- $3.x.3 altitude grid
         iwarn = go_section(3, i_diag+1, 3, 0, unit=unit)
         tmp = fltarr(n_grid)
         readf, unit, tmp
         diagnostics[i_to_read].altitude = ptr_new(tmp)

;--- $3.x.5 estimate of total error
         iwarn = go_section(3, i_diag+1, 5, 0, unit=unit)
         tmp = fltarr(n_grid)
         readf, unit, tmp
         diagnostics[i_to_read].total_error = ptr_new(tmp)

;--- $3.x.6 random error covariance matrix
         iwarn = go_section(3, i_diag+1, 6, 0, unit=unit)
         tmp = fltarr(n_grid,n_grid)
         readf, unit, tmp
         diagnostics[i_to_read].random_error_covariance = ptr_new(tmp)

;--- $3.x.7 averaging kernel matrix
         iwarn = go_section(3, i_diag+1, 7, 0, unit=unit)
         tmp = fltarr(n_grid,n_grid)
         readf, unit, tmp
         diagnostics[i_to_read].ak_matrix = ptr_new(tmp)

;--- increase array index by one
         i_to_read = i_to_read + 1
         
      endfor
      
      if(i_to_read ne n_to_read) then stop

      free_lun, unit

;--- map the indices to ease access
      return, map_kadex_diagnostics(result, diagnostics)

   endif

;--- no diagnostics: define diagnostics to avoid assignment problems
   diagnostics = 0

   free_lun, unit
   return, 0

IO_ERROR:
   if(n_elements(unit) ne 0) then free_lun, unit
   print, routine+': unspecified IO-error occured while reading file '+file
   return, 1

end

;==============================================================================
function read_kadex, d, data_file

   if(n_params() lt 1 or n_params() gt 2) then begin
      print, ' usage: success = read_kadex(d, data_file)'
      return, 1
   endif
   
;--- no paramter file available or file is no string: pick file from browser
   if(n_elements(data_file) ne 1) then begin
      data_file = dialog_pickfile(/read)
   endif else if((size(data_file, /str)).type_name ne 'STRING') then begin
      data_file = dialog_pickfile(/read)
   endif

;--- read header info of data file, result is stored in
;    a structure called header
   if(read_kadex_header(header, data_file) ne 0) then return, 1

;--- read the corresponding result data into a specific data structure.
;    We have pre-selected data wrt latitude and solar zenith angle,
;    this selection is represented by index and avoids reading of
;    unwanted data.
   if(read_kadex_result(result, data_file, index=index) ne 0) then return, 1

;--- Result contains a subset of all data contained in the data file.
;    To get only the necessar
;    read the relevant diagnostics data; in a first step prepare a
;    list of needed diagnostics indices
   if(read_kadex_diagnostics(diagnostics, result, data_file) ne 0) then return, 1

   d = {header: header, result: result, diagnostics: diagnostics}

   return, 0

end
